package com.sun.xml.internal.ws.encoding;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import com.sun.xml.internal.ws.streaming.XMLStreamReaderUtil;
import com.sun.xml.internal.ws.util.xml.NamespaceContextExAdaper;
import com.sun.xml.internal.org.jvnet.staxex.XMLStreamReaderEx;
import com.sun.xml.internal.ws.util.xml.XMLStreamReaderFilter;
import javax.xml.namespace.NamespaceContext;
import com.sun.xml.internal.org.jvnet.staxex.NamespaceContextEx;
import com.sun.xml.internal.ws.streaming.XMLStreamWriterUtil;
import javax.xml.bind.attachment.AttachmentMarshaller;
import java.util.Map;
import com.sun.xml.internal.org.jvnet.staxex.Base64Data;
import com.sun.xml.internal.bind.DatatypeConverterImpl;
import com.sun.xml.internal.ws.util.ByteArrayDataSource;
import javax.xml.stream.XMLStreamWriter;
import com.sun.xml.internal.ws.streaming.MtomStreamWriter;
import com.sun.xml.internal.org.jvnet.staxex.XMLStreamWriterEx;
import com.sun.xml.internal.ws.util.xml.XMLStreamWriterFilter;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.ws.api.pipe.Codec;
import java.io.InputStream;
import java.nio.channels.ReadableByteChannel;
import javax.xml.stream.XMLStreamReader;
import com.sun.xml.internal.ws.server.UnsupportedMediaException;
import com.sun.xml.internal.ws.message.MimeAttachmentSet;
import com.sun.xml.internal.ws.api.streaming.XMLStreamReaderFactory;
import java.nio.charset.Charset;
import java.nio.channels.WritableByteChannel;
import javax.activation.DataHandler;
import com.sun.xml.internal.ws.developer.StreamingDataHandler;
import com.sun.xml.internal.ws.api.message.Attachment;
import com.sun.xml.internal.ws.api.message.AttachmentSet;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import javax.xml.stream.XMLStreamException;
import javax.xml.ws.WebServiceException;
import com.sun.xml.internal.ws.api.streaming.XMLStreamWriterFactory;
import java.util.ArrayList;
import java.io.OutputStream;
import java.util.UUID;
import com.sun.xml.internal.ws.api.pipe.ContentType;
import com.sun.xml.internal.ws.api.message.Packet;
import com.sun.xml.internal.ws.api.WSFeatureList;
import com.sun.xml.internal.ws.api.SOAPVersion;
import com.sun.xml.internal.ws.developer.SerializationFeature;
import javax.xml.ws.soap.MTOMFeature;
import com.sun.xml.internal.ws.api.pipe.StreamSOAPCodec;

public class MtomCodec extends MimeCodec
{
    public static final String XOP_XML_MIME_TYPE = "application/xop+xml";
    public static final String XOP_LOCALNAME = "Include";
    public static final String XOP_NAMESPACEURI = "http://www.w3.org/2004/08/xop/include";
    private final StreamSOAPCodec codec;
    private final MTOMFeature mtomFeature;
    private final SerializationFeature sf;
    private static final String DECODED_MESSAGE_CHARSET = "decodedMessageCharset";
    
    MtomCodec(SOAPVersion version, StreamSOAPCodec codec, WSFeatureList features) {
        super(version, features);
        this.codec = codec;
        this.sf = (SerializationFeature)features.get(SerializationFeature.class);
        MTOMFeature mtom = (MTOMFeature)features.get(MTOMFeature.class);
        if (mtom == null) {
            this.mtomFeature = new MTOMFeature();
        }
        else {
            this.mtomFeature = mtom;
        }
    }
    
    @Override
    public ContentType getStaticContentType(Packet packet) {
        return getStaticContentTypeStatic(packet, this.version);
    }
    
    public static ContentType getStaticContentTypeStatic(Packet packet, SOAPVersion version) {
        ContentType ct = (ContentType)packet.getInternalContentType();
        if (ct != null) {
            return ct;
        }
        String uuid = UUID.randomUUID().toString();
        String boundary = "uuid:" + uuid;
        String rootId = "<rootpart*" + uuid + "@example.jaxws.sun.com>";
        String soapActionParameter = SOAPVersion.SOAP_11.equals(version) ? null : createActionParameter(packet);
        String boundaryParameter = "boundary=\"" + boundary + "\"";
        String messageContentType = "multipart/related;start=\"" + rootId + "\";type=\"" + "application/xop+xml" + "\";" + boundaryParameter + ";start-info=\"" + version.contentType + ((soapActionParameter == null) ? "" : soapActionParameter) + "\"";
        ContentTypeImpl ctImpl = SOAPVersion.SOAP_11.equals(version) ? new ContentTypeImpl(messageContentType, (packet.soapAction == null) ? "" : packet.soapAction, null) : new ContentTypeImpl(messageContentType, null, null);
        ctImpl.setBoundary(boundary);
        ctImpl.setRootId(rootId);
        packet.setContentType(ctImpl);
        return ctImpl;
    }
    
    private static String createActionParameter(Packet packet) {
        return (packet.soapAction != null) ? (";action=\\\"" + packet.soapAction + "\\\"") : "";
    }
    
    @Override
    public ContentType encode(Packet packet, OutputStream out) throws IOException {
        ContentTypeImpl ctImpl = (ContentTypeImpl)this.getStaticContentType(packet);
        String boundary = ctImpl.getBoundary();
        String rootId = ctImpl.getRootId();
        if (packet.getMessage() != null) {
            try {
                String encoding = this.getPacketEncoding(packet);
                packet.invocationProperties.remove("decodedMessageCharset");
                String actionParameter = getActionParameter(packet, this.version);
                String soapXopContentType = getSOAPXopContentType(encoding, this.version, actionParameter);
                MimeCodec.writeln("--" + boundary, out);
                writeMimeHeaders(soapXopContentType, rootId, out);
                List<MtomCodec.ByteArrayBuffer> mtomAttachments = new ArrayList();
                MtomCodec.MtomStreamWriterImpl writer = new MtomCodec.MtomStreamWriterImpl(XMLStreamWriterFactory.create(out, encoding), mtomAttachments, boundary, this.mtomFeature);
                packet.getMessage().writeTo(writer);
                XMLStreamWriterFactory.recycle(writer);
                MimeCodec.writeln(out);
                for (MtomCodec.ByteArrayBuffer bos : mtomAttachments) {
                    bos.write(out);
                }
                this.writeNonMtomAttachments(packet.getMessage().getAttachments(), out, boundary);
                MimeCodec.writeAsAscii("--" + boundary, out);
                MimeCodec.writeAsAscii("--", out);
            }
            catch (XMLStreamException e) {
                throw new WebServiceException(e);
            }
        }
        return ctImpl;
    }
    
    public static String getSOAPXopContentType(String encoding, SOAPVersion version, String actionParameter) {
        return "application/xop+xml;charset=" + encoding + ";type=\"" + version.contentType + actionParameter + "\"";
    }
    
    public static String getActionParameter(Packet packet, SOAPVersion version) {
        return (version == SOAPVersion.SOAP_11) ? "" : createActionParameter(packet);
    }
    
    public static void writeMimeHeaders(String contentType, String contentId, OutputStream out) throws IOException {
        String cid = contentId;
        if (cid != null && cid.length() > 0 && cid.charAt(0) != '<') {
            cid = 60 + cid + 62;
        }
        MimeCodec.writeln("Content-Id: " + cid, out);
        MimeCodec.writeln("Content-Type: " + contentType, out);
        MimeCodec.writeln("Content-Transfer-Encoding: binary", out);
        MimeCodec.writeln(out);
    }
    
    private void writeNonMtomAttachments(AttachmentSet attachments, OutputStream out, String boundary) throws IOException {
        for (Attachment att : attachments) {
            DataHandler dh = att.asDataHandler();
            if ((dh instanceof StreamingDataHandler)) {
                StreamingDataHandler sdh = (StreamingDataHandler)dh;
                if (sdh.getHrefCid() != null) {
                    continue;
                }
            }
            MimeCodec.writeln("--" + boundary, out);
            writeMimeHeaders(att.getContentType(), att.getContentId(), out);
            att.writeTo(out);
            MimeCodec.writeln(out);
        }
    }
    
    public ContentType encode(Packet packet, WritableByteChannel buffer) {
        throw new UnsupportedOperationException();
    }
    
    @Override
    public MtomCodec copy() {
        return new MtomCodec(this.version, (StreamSOAPCodec)this.codec.copy(), this.features);
    }
    
    private static String encodeCid() {
        String cid = "example.jaxws.sun.com";
        String name = UUID.randomUUID() + "@";
        return name + cid;
    }
    
    @Override
    protected void decode(MimeMultipartParser mpp, Packet packet) throws IOException {
        String charset = null;
        String ct = mpp.getRootPart().getContentType();
        if (ct != null) {
            charset = new ContentTypeImpl(ct).getCharSet();
        }
        if (charset != null && !Charset.isSupported(charset)) {
            throw new UnsupportedMediaException(charset);
        }
        if (charset != null) {
            packet.invocationProperties.put("decodedMessageCharset", charset);
        }
        else {
            packet.invocationProperties.remove("decodedMessageCharset");
        }
        XMLStreamReader mtomReader = new MtomCodec.MtomXMLStreamReaderEx(mpp, XMLStreamReaderFactory.create(null, mpp.getRootPart().asInputStream(), charset, true));
        packet.setMessage(this.codec.decode(mtomReader, new MimeAttachmentSet(mpp)));
        packet.setMtomFeature(this.mtomFeature);
        packet.setContentType(mpp.getContentType());
        return;
    }
    
    private String getPacketEncoding(Packet packet) {
        if (this.sf != null && this.sf.getEncoding() != null) {
            return this.sf.getEncoding().equals("") ? "utf-8" : this.sf.getEncoding();
        }
        return determinePacketEncoding(packet);
    }
    
    public static String determinePacketEncoding(Packet packet) {
        if (packet != null && packet.endpoint != null) {
            String charset = (String)packet.invocationProperties.get("decodedMessageCharset");
            return (charset == null) ? "utf-8" : charset;
        }
        return "utf-8";
    }
    
    static String access$000() {
        return encodeCid();
    }
    
    public static class ByteArrayBuffer
    {
        final String contentId;
        private final DataHandler dh;
        private final String boundary;
        
        ByteArrayBuffer(@NotNull DataHandler dh, String b) {
            this.dh = dh;
            String cid = null;
            if ((dh instanceof StreamingDataHandler)) {
                StreamingDataHandler sdh = (StreamingDataHandler)dh;
                if (sdh.getHrefCid() != null) {
                    cid = sdh.getHrefCid();
                }
            }
            this.contentId = ((cid != null) ? cid : MtomCodec.access$000());
            this.boundary = b;
        }
        
        public void write(OutputStream os) throws IOException {
            MimeCodec.writeln("--" + this.boundary, os);
            MtomCodec.writeMimeHeaders(this.dh.getContentType(), this.contentId, os);
            this.dh.writeTo(os);
            MimeCodec.writeln(os);
        }
    }
    
    public static class MtomStreamWriterImpl extends XMLStreamWriterFilter implements XMLStreamWriterEx, MtomStreamWriter, HasEncoding
    {
        private final List<MtomCodec.ByteArrayBuffer> mtomAttachments;
        private final String boundary;
        private final MTOMFeature myMtomFeature;
        
        public MtomStreamWriterImpl(XMLStreamWriter w, List<MtomCodec.ByteArrayBuffer> mtomAttachments, String b, MTOMFeature myMtomFeature) {
            super(w);
            this.mtomAttachments = mtomAttachments;
            this.boundary = b;
            this.myMtomFeature = myMtomFeature;
        }
        
        public void writeBinary(byte[] data, int start, int len, String contentType) throws XMLStreamException {
            if (this.myMtomFeature.getThreshold() > len) {
                this.writeCharacters(DatatypeConverterImpl._printBase64Binary(data, start, len));
                return;
            }
            MtomCodec.ByteArrayBuffer bab = new MtomCodec.ByteArrayBuffer(new DataHandler(new ByteArrayDataSource(data, start, len, contentType)), this.boundary);
            this.writeBinary(bab);
            return;
        }
        
        public void writeBinary(DataHandler dataHandler) throws XMLStreamException {
            this.writeBinary(new MtomCodec.ByteArrayBuffer(dataHandler, this.boundary));
        }
        
        public OutputStream writeBinary(String contentType) throws XMLStreamException {
            throw new UnsupportedOperationException();
        }
        
        public void writePCDATA(CharSequence data) throws XMLStreamException {
            if (data == null) {
                return;
            }
            if ((data instanceof Base64Data)) {
                Base64Data binaryData = (Base64Data)data;
                this.writeBinary(binaryData.getDataHandler());
                return;
            }
            this.writeCharacters(data.toString());
            return;
        }
        
        private void writeBinary(MtomCodec.ByteArrayBuffer bab) {
            try {
                this.mtomAttachments.add(bab);
                String prefix = this.writer.getPrefix("http://www.w3.org/2004/08/xop/include");
                if (prefix == null || !prefix.equals("xop")) {
                    this.writer.setPrefix("xop", "http://www.w3.org/2004/08/xop/include");
                    this.writer.writeNamespace("xop", "http://www.w3.org/2004/08/xop/include");
                }
                this.writer.writeStartElement("http://www.w3.org/2004/08/xop/include", "Include");
                this.writer.writeAttribute("href", "cid:" + bab.contentId);
                this.writer.writeEndElement();
                this.writer.flush();
            }
            catch (XMLStreamException e) {
                throw new WebServiceException(e);
            }
        }
        
        public Object getProperty(String name) throws IllegalArgumentException {
            if (name.equals("sjsxp-outputstream") && this.writer instanceof Map) {
                Object obj = ((Map)this.writer).get("sjsxp-outputstream");
                if (obj != null) {
                    return obj;
                }
            }
            return super.getProperty(name);
        }
        
        public AttachmentMarshaller getAttachmentMarshaller() {
            return new AttachmentMarshaller() {
                public String addMtomAttachment(DataHandler data, String elementNamespace, String elementLocalName) {
                    MtomCodec.ByteArrayBuffer bab = new MtomCodec.ByteArrayBuffer(data, MtomCodec.MtomStreamWriterImpl.this.boundary);
                    MtomCodec.MtomStreamWriterImpl.this.mtomAttachments.add(bab);
                    return "cid:" + bab.contentId;
                }
                
                public String addMtomAttachment(byte[] data, int offset, int length, String mimeType, String elementNamespace, String elementLocalName) {
                    if (MtomCodec.MtomStreamWriterImpl.this.myMtomFeature.getThreshold() > length) {
                        return null;
                    }
                    MtomCodec.ByteArrayBuffer bab = new MtomCodec.ByteArrayBuffer(new DataHandler(new ByteArrayDataSource(data, offset, length, mimeType)), MtomCodec.MtomStreamWriterImpl.this.boundary);
                    MtomCodec.MtomStreamWriterImpl.this.mtomAttachments.add(bab);
                    return "cid:" + bab.contentId;
                }
                
                public String addSwaRefAttachment(DataHandler data) {
                    MtomCodec.ByteArrayBuffer bab = new MtomCodec.ByteArrayBuffer(data, MtomCodec.MtomStreamWriterImpl.this.boundary);
                    MtomCodec.MtomStreamWriterImpl.this.mtomAttachments.add(bab);
                    return "cid:" + bab.contentId;
                }
                
                public boolean isXOPPackage() {
                    return true;
                }
            };
        }
        
        public List<MtomCodec.ByteArrayBuffer> getMtomAttachments() {
            return this.mtomAttachments;
        }
        
        public String getEncoding() {
            return XMLStreamWriterUtil.getEncoding(this.writer);
        }
        
        public NamespaceContextEx getNamespaceContext() {
            NamespaceContext nsContext = this.writer.getNamespaceContext();
            return new MtomNamespaceContextEx(nsContext);
        }
        
        private static class MtomNamespaceContextEx implements NamespaceContextEx
        {
            private final NamespaceContext nsContext;
            
            public MtomNamespaceContextEx(NamespaceContext nsContext) {
                this.nsContext = nsContext;
            }
            
            public Iterator<NamespaceContextEx.Binding> iterator() {
                throw new UnsupportedOperationException();
            }
            
            public String getNamespaceURI(String prefix) {
                return this.nsContext.getNamespaceURI(prefix);
            }
            
            public String getPrefix(String namespaceURI) {
                return this.nsContext.getPrefix(namespaceURI);
            }
            
            public Iterator getPrefixes(String namespaceURI) {
                return this.nsContext.getPrefixes(namespaceURI);
            }
        }
    }
    
    public static class MtomXMLStreamReaderEx extends XMLStreamReaderFilter implements XMLStreamReaderEx
    {
        private final MimeMultipartParser mimeMP;
        private boolean xopReferencePresent;
        private Base64Data base64AttData;
        private char[] base64EncodedText;
        private String xopHref;
        
        public MtomXMLStreamReaderEx(MimeMultipartParser mimeMP, XMLStreamReader reader) {
            super(reader);
            this.xopReferencePresent = false;
            this.mimeMP = mimeMP;
        }
        
        public CharSequence getPCDATA() throws XMLStreamException {
            if (this.xopReferencePresent) {
                return this.base64AttData;
            }
            return this.reader.getText();
        }
        
        public NamespaceContextEx getNamespaceContext() {
            return new NamespaceContextExAdaper(this.reader.getNamespaceContext());
        }
        
        public String getElementTextTrim() throws XMLStreamException {
            throw new UnsupportedOperationException();
        }
        
        public int getTextLength() {
            if (this.xopReferencePresent) {
                return this.base64AttData.length();
            }
            return this.reader.getTextLength();
        }
        
        public int getTextStart() {
            if (this.xopReferencePresent) {
                return 0;
            }
            return this.reader.getTextStart();
        }
        
        public int getEventType() {
            if (this.xopReferencePresent) {
                return 4;
            }
            return super.getEventType();
        }
        
        public int next() throws XMLStreamException {
            int event = this.reader.next();
            if (event == 1 && this.reader.getLocalName().equals("Include") && this.reader.getNamespaceURI().equals("http://www.w3.org/2004/08/xop/include")) {
                String href = this.reader.getAttributeValue(null, "href");
                try {
                    this.xopHref = href;
                    Attachment att = this.getAttachment(href);
                    if (att != null) {
                        DataHandler dh = att.asDataHandler();
                        if ((dh instanceof StreamingDataHandler)) {
                            ((StreamingDataHandler)dh).setHrefCid(att.getContentId());
                        }
                        (this.base64AttData = new Base64Data()).set(dh);
                    }
                    this.xopReferencePresent = true;
                }
                catch (IOException e) {
                    throw new WebServiceException(e);
                }
                XMLStreamReaderUtil.nextElementContent(this.reader);
                return 4;
            }
            if (this.xopReferencePresent) {
                this.xopReferencePresent = false;
                this.base64EncodedText = null;
                this.xopHref = null;
            }
            return event;
        }
        
        private String decodeCid(String cid) {
            try {
                cid = URLDecoder.decode(cid, "utf-8");
            }
            catch (UnsupportedEncodingException ex) {}
            return cid;
        }
        
        private Attachment getAttachment(String cid) throws IOException {
            if (cid.startsWith("cid:")) {
                cid = cid.substring(4, cid.length());
            }
            if (cid.indexOf(37) != -1) {
                cid = this.decodeCid(cid);
                return this.mimeMP.getAttachmentPart(cid);
            }
            return this.mimeMP.getAttachmentPart(cid);
        }
        
        public char[] getTextCharacters() {
            if (this.xopReferencePresent) {
                char[] chars = new char[this.base64AttData.length()];
                this.base64AttData.writeTo(chars, 0);
                return chars;
            }
            return this.reader.getTextCharacters();
        }
        
        public int getTextCharacters(int sourceStart, char[] target, int targetStart, int length) throws XMLStreamException {
            if (this.xopReferencePresent) {
                if (target == null) {
                    throw new NullPointerException("target char array can't be null");
                }
                if (targetStart < 0 || length < 0 || sourceStart < 0 || targetStart >= target.length || targetStart + length > target.length) {
                    throw new IndexOutOfBoundsException();
                }
                int textLength = this.base64AttData.length();
                if (sourceStart > textLength) {
                    throw new IndexOutOfBoundsException();
                }
                if (this.base64EncodedText == null) {
                    this.base64EncodedText = new char[this.base64AttData.length()];
                    this.base64AttData.writeTo(this.base64EncodedText, 0);
                }
                int copiedLength = Math.min(textLength - sourceStart, length);
                System.arraycopy(this.base64EncodedText, sourceStart, target, targetStart, copiedLength);
                return copiedLength;
            }
            else {
                return this.reader.getTextCharacters(sourceStart, target, targetStart, length);
            }
        }
        
        public String getText() {
            if (this.xopReferencePresent) {
                return this.base64AttData.toString();
            }
            return this.reader.getText();
        }
        
        protected boolean isXopReference() throws XMLStreamException {
            return this.xopReferencePresent;
        }
        
        protected String getXopHref() {
            return this.xopHref;
        }
        
        public MimeMultipartParser getMimeMultipartParser() {
            return this.mimeMP;
        }
    }
}
